﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;

namespace de.mhoapps.DropToZip.View
{
    /// <summary>
    /// Eine Basisklasse zum Ausführen einer beliebigen Animationsmethode, wenn ein Boolescher Wert auf true gesetzt ist
    /// und eine umgekehrte Animation, wenn auf false gesetzt
    /// </ summary>
    /// <typeparam name = "Parent"> </ typeparam>
    public abstract class AnimateBaseProperty<Parent> : AttachedPropertyBase<Parent, bool>
        where Parent : AttachedPropertyBase<Parent, bool>, new()
    {
        #region Protected Properties

        /// <summary>
        /// True, wenn dies das allererste Mal ist, dass der Wert aktualisiert wurde
        /// Wird verwendet, um sicherzustellen, dass die Logik beim ersten Laden mindestens einmal ausgeführt wird
        /// </ summary>
        protected Dictionary<WeakReference, bool> _alreadyLoaded = new Dictionary<WeakReference, bool>();

        /// <summary>
        /// Der zuletzt verwendete Wert, wenn wir vor dem ersten Laden einen geänderten Wert erhalten
        /// </ summary>
        protected Dictionary<WeakReference, bool> _firstLoadValue = new Dictionary<WeakReference, bool>();

        #endregion

        public override void OnValueUpdated(DependencyObject sender, object value)
        {
            if (!(sender is FrameworkElement element)) return;

            var alreadyLoadedReference = _alreadyLoaded.FirstOrDefault(f => f.Key.Target == sender);

            var firstLoadReference = _firstLoadValue.FirstOrDefault(f => f.Key.Target == sender);

            if ((bool)sender.GetValue(ValueProperty) == (bool)value && alreadyLoadedReference.Key != null) return;

            if (alreadyLoadedReference.Key == null)
            {
                var weakReference = new WeakReference(sender);

                _alreadyLoaded[weakReference] = false;

                element.Visibility = Visibility.Hidden;

                RoutedEventHandler onLoaded = null;
                onLoaded = async (ss, ee) =>
                {
                    element.Loaded -= onLoaded;
                    await Task.Delay(5);

                    firstLoadReference = _firstLoadValue.FirstOrDefault(f => f.Key.Target == sender);

                    DoAnimation(element, firstLoadReference.Key != null ? firstLoadReference.Value : (bool)value, true);

                    _alreadyLoaded[weakReference] = true;
                };

                element.Loaded += onLoaded;
            }

            else if (alreadyLoadedReference.Value == false)
                _firstLoadValue[new WeakReference(sender)] = (bool)value;
            else
                DoAnimation(element, (bool)value, false);
        }

        /// <summary>
        /// Die Animationsmethode, die ausgelöst wird, wenn sich der Wert ändert
        /// </ summary>
        /// <param name = "element"> Das Element </ param>
        /// <param name = "value"> Der neue Wert </ param>
        /// <param name="firstLoad">boolscher Ausdruck ob erstes laden</param>
        protected virtual void DoAnimation(FrameworkElement element, bool value, bool firstLoad) { }
    }

    /// <summary>
    /// Blendet ein Bild ein, sobald sich die Quelle ändert
    /// </ summary>
    public class FadeInImageOnLoadProperty : AnimateBaseProperty<FadeInImageOnLoadProperty>
    {
        public override void OnValueUpdated(DependencyObject sender, object value)
        {
            if (!(sender is Image image)) return;

            if ((bool)value)
                image.TargetUpdated += Image_TargetUpdatedAsync;
            else
                image.TargetUpdated -= Image_TargetUpdatedAsync;
        }

        private async void Image_TargetUpdatedAsync(object sender, System.Windows.Data.DataTransferEventArgs e)
        {
            await (sender as Image).FadeInAsync(false);
        }
    }

    /// <summary>
    /// Animiert ein Framework-Element, indem es in der Anzeige von links eingeschoben wird
    /// und auf hide nach links rausrutschen
    /// </ summary>
    public class AnimateSlideInFromLeftProperty : AnimateBaseProperty<AnimateSlideInFromLeftProperty>
    {
        protected override async void DoAnimation(FrameworkElement element, bool value, bool firstLoad)
        {
            if (value)
                await element.SlideAndFadeInAsync(AnimationSlideInDirection.Left, firstLoad, firstLoad ? 0 : 0.3f, keepMargin: false);
            else
                await element.SlideAndFadeOutAsync(AnimationSlideInDirection.Left, firstLoad ? 0 : 0.3f, keepMargin: false);
        }
    }

    /// <summary>
    /// Animiert ein Framework-Element, indem es von rechts eingeschoben wird
    /// und auf hide nach rechts rausrutschen
    /// </ summary>
    public class AnimateSlideInFromRightProperty : AnimateBaseProperty<AnimateSlideInFromRightProperty>
    {
        protected override async void DoAnimation(FrameworkElement element, bool value, bool firstLoad)
        {
            if (value)
                await element.SlideAndFadeInAsync(AnimationSlideInDirection.Right, firstLoad, firstLoad ? 0 : 0.3f, keepMargin: false);
            else
                await element.SlideAndFadeOutAsync(AnimationSlideInDirection.Right, firstLoad ? 0 : 0.3f, keepMargin: false);
        }
    }

    /// <summary>
    /// Animiert ein Framework-Element, indem es von rechts eingeschoben wird
    /// und auf hide nach rechts rausrutschen
    /// HINWEIS: Der Rand bleibt erhalten
    /// </ summary>
    public class AnimateSlideInFromRightMarginProperty : AnimateBaseProperty<AnimateSlideInFromRightMarginProperty>
    {
        protected override async void DoAnimation(FrameworkElement element, bool value, bool firstLoad)
        {
            if (value)
                await element.SlideAndFadeInAsync(AnimationSlideInDirection.Right, firstLoad, firstLoad ? 0 : 0.3f, keepMargin: true);
            else
                await element.SlideAndFadeOutAsync(AnimationSlideInDirection.Right, firstLoad ? 0 : 0.3f, keepMargin: true);
        }
    }

    /// <summary>
    /// Animiert ein Framework-Element, das bei der Anzeige von oben nach unten gleitet
    /// und auf hide nach oben rausrutschen
    /// </ summary>
    public class AnimateSlideInFromTopProperty : AnimateBaseProperty<AnimateSlideInFromTopProperty>
    {
        protected override async void DoAnimation(FrameworkElement element, bool value, bool firstLoad)
        {
            if (value)
                await element.SlideAndFadeInAsync(AnimationSlideInDirection.Top, firstLoad, firstLoad ? 0 : 0.3f, keepMargin: false);
            else
                await element.SlideAndFadeOutAsync(AnimationSlideInDirection.Top, firstLoad ? 0 : 0.3f, keepMargin: false);
        }
    }

    /// <summary>
    /// Animiert ein Framework-Element, das bei der Anzeige von unten nach oben gleitet
    /// und auf hide nach unten rausrutschen
    /// </ summary>
    public class AnimateSlideInFromBottomProperty : AnimateBaseProperty<AnimateSlideInFromBottomProperty>
    {
        protected override async void DoAnimation(FrameworkElement element, bool value, bool firstLoad)
        {
            if (value)
                await element.SlideAndFadeInAsync(AnimationSlideInDirection.Bottom, firstLoad, firstLoad ? 0 : 0.3f, keepMargin: false);
            else
                await element.SlideAndFadeOutAsync(AnimationSlideInDirection.Bottom, firstLoad ? 0 : 0.3f, keepMargin: false);
        }
    }

    /// <summary>
    /// Animiert ein Framework-Element, das beim Laden von unten nach oben rutscht
    /// wenn der Wert wahr ist
    /// </ summary>
    public class AnimateSlideInFromBottomOnLoadProperty : AnimateBaseProperty<AnimateSlideInFromBottomOnLoadProperty>
    {
        protected override async void DoAnimation(FrameworkElement element, bool value, bool firstLoad)
        {
            await element.SlideAndFadeInAsync(AnimationSlideInDirection.Bottom, !value, !value ? 0 : 0.3f, keepMargin: false);
        }
    }

    /// <summary>
    /// Animiert ein Framework-Element, das bei der Anzeige von unten nach oben gleitet
    /// und auf hide nach unten rausrutschen
    /// HINWEIS: Der Rand bleibt erhalten
    /// </ summary>
    public class AnimateSlideInFromBottomMarginProperty : AnimateBaseProperty<AnimateSlideInFromBottomMarginProperty>
    {
        protected override async void DoAnimation(FrameworkElement element, bool value, bool firstLoad)
        {
            if (value)
                await element.SlideAndFadeInAsync(AnimationSlideInDirection.Bottom, firstLoad, firstLoad ? 0 : 0.3f, keepMargin: true);
            else
                await element.SlideAndFadeOutAsync(AnimationSlideInDirection.Bottom, firstLoad ? 0 : 0.3f, keepMargin: true);
        }
    }

    /// <summary>
    /// Animiert ein Framework-Element, das bei .Show eingeblendet wird
    /// und ausgeblendet bei .Hide
    /// </ summary>
    public class AnimateFadeInProperty : AnimateBaseProperty<AnimateFadeInProperty>
    {
        protected override async void DoAnimation(FrameworkElement element, bool value, bool firstLoad)
        {
            if (value)
                await element.FadeInAsync(firstLoad, firstLoad ? 0 : 0.3f);
            else
                await element.FadeOutAsync(firstLoad ? 0 : 0.3f);
        }
    }
}
