﻿using de.mhoapps.DropToZip.ViewModel;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Security;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;

namespace de.mhoapps.DropToZip.View
{
    /// <summary>
    /// Erstellt eine bindbare angehängte Eigenschaft für die Eigenschaft <see cref = "PasswordBox.SecurePassword" />.
    /// </ summary>
    public static class SecurePasswordAttachedProperty
    {
        // Ein angefügtes Verhalten funktioniert nicht, da bei der Modellvalidierung nicht das richtige Steuerelement ausgewählt wurde
        public static readonly DependencyProperty SecurePasswordBindingProperty = DependencyProperty.RegisterAttached(
            "SecurePassword",
            typeof(SecureString),
            typeof(SecurePasswordAttachedProperty),
            new FrameworkPropertyMetadata(new SecureString(), FrameworkPropertyMetadataOptions.BindsTwoWayByDefault, AttachedPropertyValueChanged)
        );

        private static readonly DependencyProperty _passwordBindingMarshallerProperty = DependencyProperty.RegisterAttached(
            "PasswordBindingMarshaller",
            typeof(PasswordBindingMarshaller),
            typeof(SecurePasswordAttachedProperty),
            new PropertyMetadata()
        );

        public static void SetSecurePassword(PasswordBox element, SecureString secureString)
        {
            element.SetValue(SecurePasswordBindingProperty, secureString);
        }

        public static SecureString GetSecurePassword(PasswordBox element)
        {
            return element.GetValue(SecurePasswordBindingProperty) as SecureString;
        }

        private static void AttachedPropertyValueChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            // Wir müssen uns mit einem der Ereignisse des Elements verbinden
            // Damit der GC das Steuerelement erfassen kann, wird der Ereignishandler in ein Objekt eingeschlossen, das sich in einer angehängten Eigenschaft befindet
            // Versuchen Sie nicht, das Unloaded-Ereignis zu verwenden, da es auch dann ausgelöst wird, wenn das Steuerelement noch aktiv ist (z. B. Tabulatoren in einem Tabulatorsteuerelement wechseln).
            var passwordBox = (PasswordBox)d;
            var bindingMarshaller = passwordBox.GetValue(_passwordBindingMarshallerProperty) as PasswordBindingMarshaller;
            if (bindingMarshaller == null)
            {
                bindingMarshaller = new PasswordBindingMarshaller(passwordBox);
                passwordBox.SetValue(_passwordBindingMarshallerProperty, bindingMarshaller);
            }

            bindingMarshaller.UpdatePasswordBox(e.NewValue as SecureString);
        }

        /// <summary>
        /// Gekapselte Ereignislogik
        /// </ summary>
        private class PasswordBindingMarshaller
        {
            private readonly PasswordBox _passwordBox;
            private bool _isMarshalling;

            public PasswordBindingMarshaller(PasswordBox passwordBox)
            {
                _passwordBox = passwordBox;
                _passwordBox.PasswordChanged += this.PasswordBoxPasswordChanged;
            }

            public void UpdatePasswordBox(SecureString newPassword)
            {
                if (_isMarshalling)
                {
                    return;
                }

                _isMarshalling = true;
                try
                {
                    // Das Einrichten von SecuredPassword löst keine visuelle Aktualisierung aus, daher müssen wir die Password-Eigenschaft verwenden
                    _passwordBox.Password = newPassword.Unsecure();

                    // Sie können die folgende Anweisung ausprobieren, die Vorteile sind jedoch in Bezug auf die Sicherheit minimal (Sie müssen das ungesicherte Kennwort zum Kopieren noch extrahieren).
                    //newPassword.CopyInto(_passwordBox.SecurePassword);
                }
                finally
                {
                    _isMarshalling = false;
                }
            }

            private void PasswordBoxPasswordChanged(object sender, RoutedEventArgs e)
            {
                if (_isMarshalling) return;

                _isMarshalling = true;
                try
                {
                    SetSecurePassword(_passwordBox, _passwordBox.SecurePassword.Copy());
                }
                finally
                {
                    _isMarshalling = false;
                }
            }
        }
    }
}
