﻿using de.mhoapps.DropToZip.ViewModel;
using System.Windows;
using System.Windows.Input;

namespace de.mhoapps.DropToZip.View
{
    /// <summary>
    /// Das ViewModel für das benutzerdefinierte flache Fenster
    /// </ summary>
    public class WindowViewModel : ViewModelBase
    {


        #region Private Member

        /// <summary>
        /// das Fenster welches dieses ViewModel kontrolliert
        /// </summary>
        private Window _window;

        /// <summary>
        /// Der <see cref="WindowResizer"/>-Helfer, der die Fenstergröße in verschiedenen Zuständen korrekt hält
        /// </ summary>
        private WindowResizer _windowResizer;

        /// <summary>
        /// Der Rand um das Fenster, um einen Schlagschatten zu ermöglichen
        /// </ summary>
        private Thickness _outerMarginSize = new Thickness(5);

        /// <summary>
        /// Der Radius der Fensterränder
        /// </ summary>
        private int _windowRadius = 10;

        /// <summary>
        /// Die letzte bekannte Dockposition
        /// </ summary>
        private WindowDockPosition _dockPosition = WindowDockPosition.Undocked;

        #endregion

        #region Public Properties

        /// <summary>
        /// Die kleinste Breite, die das Fenster erreichen kann
        /// </ summary>
        public double WindowMinimumWidth { get; set; } = 450;

        /// <summary>
        /// Die kleinste Höhe, die das Fenster erreichen kann
        /// </ summary>
        public double WindowMinimumHeight { get; set; } = 500;

        /// <summary>
        /// True wenn das Fenster gerade verschoben / gezogen wird
        /// </ summary>
        public bool BeingMoved { get; set; }


        /// <summary>
        /// True, wenn das Fenster randlos sein soll, weil es angedockt oder maximiert ist
        /// </ summary>
        public bool Borderless => (_window.WindowState == WindowState.Maximized || _dockPosition != WindowDockPosition.Undocked);

        /// <summary>
        /// Die Größe des <see cref="ResizeBorder"/> um das Fenster
        /// </ summary>
        public int ResizeBorder => _window.WindowState == WindowState.Maximized ? 0 : 4;

        /// <summary>
        /// Die Größe des <see cref="ResizeBorder"/> um das Fenster unter Berücksichtigung des äußeren Rands
        /// </ summary>
        public Thickness ResizeBorderThickness => new Thickness(OuterMarginSize.Left + ResizeBorder,
                                                                OuterMarginSize.Top + ResizeBorder,
                                                                OuterMarginSize.Right + ResizeBorder,
                                                                OuterMarginSize.Bottom + ResizeBorder);

        /// <summary>
        /// Das Auffüllen des inneren Inhalts des Hauptfensters
        /// </ summary>
        public Thickness InnerContentPadding { get; set; } = new Thickness(0);

        /// <summary>
        /// Der Rand um das Fenster, um einen Schlagschatten zu ermöglichen
        /// </ summary>
        public Thickness OuterMarginSize
        {
            get => _window.WindowState == WindowState.Maximized ? _windowResizer.CurrentMonitorMargin : (Borderless ? new Thickness(0) : _outerMarginSize);
            set => _outerMarginSize = value;
        }

        /// <summary>
        /// Der Radius der Fensterränder
        /// </ summary>
        public int WindowRadius
        {
            get => Borderless ? 0 : _windowRadius;
            set => _windowRadius = value;
        }

        /// <summary>
        /// Der rechteckige Rand um das Fenster, wenn es angedockt ist
        /// </ summary>
        public int FlatBorderThickness => Borderless && _window.WindowState != WindowState.Maximized ? 1 : 0;

        /// <summary>
        /// Der Radius der Fensterränder
        /// </ summary>
        public CornerRadius WindowCornerRadius => new CornerRadius(WindowRadius);

        /// <summary>
        /// Die Höhe der Titelleiste / Beschriftung des Fensters
        /// </ summary>
        public int TitleHeight { get; set; } = 42;

        /// <summary>
        /// die Rastergröße der Titelleiste
        /// </summary>
        public GridLength TitleHeightGridLength => new GridLength(TitleHeight + ResizeBorder);

        /// <summary>
        /// True, wenn das Fenster abgeblendet sein sollte
        /// zB wenn ein Popup sichtbar ist oder das Fenster nicht fokussiert ist
        /// </ summary>
        public bool DimmableOverlayVisible { get; set; }

        /// <summary>
        /// True, wenn das Fenster über allen anderen liegen soll
        /// </summary>
        public bool TopMost { get; set; }


        #endregion

        #region Commands

        /// <summary>
        /// der Minimieren Command
        /// </summary>
        public ICommand MinimizeCommand { get; set; }

        /// <summary>
        /// der Maximieren Command
        /// </summary>
        public ICommand MaximizeCommand { get; set; }

        /// <summary>
        /// der Command zum schließen
        /// </summary>
        public ICommand CloseCommand { get; set; }

        /// <summary>
        /// der Command zum Anzeigen des SystemMenü
        /// </summary>
        public ICommand MenuCommand { get; set; }

        /// <summary>
        /// der Command zum schalten von <see cref="TopMost"/>
        /// </summary>
        public ICommand TopMostCommand { get; set; }


        #endregion

        #region Constructor

        /// <summary>
        /// Default constructor
        /// </summary>
        public WindowViewModel(Window window)
        {
            _window = window;

            _window.StateChanged += (sender, e) =>
            {
                WindowResized();
            };
            WindowTitle = "DropToZip";


            InitializeCommands();

            _windowResizer = new WindowResizer(_window);

            _windowResizer.WindowDockChanged += (dock) =>
            {
                _dockPosition = dock;

                WindowResized();
            };

            _windowResizer.WindowStartedMove += () =>
            {
                BeingMoved = true;
            };

            _windowResizer.WindowFinishedMove += () =>
            {
                BeingMoved = false;

                if (_dockPosition == WindowDockPosition.Undocked && _window.Top == _windowResizer.CurrentScreenSize.Top)
                    _window.Top = -OuterMarginSize.Top;
                WindowResized();
            };
        }

        #endregion

        #region Initialize Commands

        private void InitializeCommands()
        {
            MinimizeCommand = new RelayCommand(c => _window.WindowState = WindowState.Minimized);
            MaximizeCommand = new RelayCommand(c => _window.WindowState ^= WindowState.Maximized);
            CloseCommand = new RelayCommand(c => _window.Close());
            MenuCommand = new RelayCommand(c => SystemCommands.ShowSystemMenu(_window, GetMousePosition()));
            TopMostCommand = new RelayCommand(c => _window.Topmost ^= true);
        }

        #endregion

        #region Private Helpers

        /// <summary>
        /// Ruft die aktuelle Mausposition auf dem Bildschirm ab
        /// </ summary>
        /// <returns> </ returns>
        private Point GetMousePosition()
        {
            return _windowResizer.GetCursorPosition();
        }

        /// <summary>
        /// Wenn die Größe des Fensters auf eine bestimmte Position geändert wird (angedockt oder maximiert)
        /// Dadurch werden alle erforderlichen Eigenschaftsänderungsereignisse aktualisiert, um die Rahmen- und Radiuswerte festzulegen
        /// </ summary>
        private void WindowResized()
        {
            OnPropertyChanged(nameof(Borderless));
            OnPropertyChanged(nameof(FlatBorderThickness));
            OnPropertyChanged(nameof(ResizeBorderThickness));
            OnPropertyChanged(nameof(OuterMarginSize));
            OnPropertyChanged(nameof(WindowRadius));
            OnPropertyChanged(nameof(WindowCornerRadius));
        }
        #endregion
    }
}
