﻿using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Security;
using System.Threading.Tasks;
using System.Windows.Input;

namespace de.mhoapps.DropToZip.ViewModel
{
    /// <summary>
    /// das Hauptviewmodel
    /// </summary>
    public class MainViewModel : ViewModelBase, IFileDragDrop
    {
        #region Constructor

        /// <summary>
        /// default Constructor
        /// </summary>
        public MainViewModel()
        {
            ClearItems();
            InitializeCommands();  
        }

        #endregion

        #region Public Properties

        /// <summary>
        /// Auflistung der gedroppten Files
        /// </summary>
        public ObservableCollection<DropFileViewModel> Items { get; set; }

        /// <summary>
        /// das ausgewählte File
        /// </summary>
        public DropFileViewModel SelectedItem { get; set; }

        /// <summary>
        /// das Password für den ZipContainer
        /// </summary>
        public PasswordBoxViewModel ZipPassword { get; set; }

        /// <summary>
        /// True wenn ein Zip erstellt wedren kann
        /// Zip kann nur erstellt werden wenn UseEnxryption = false, oder UseEncryption=true UND ein Passwort angegeben wurde
        /// </summary>
        public bool CanCreateZip => UseEnryption ? (ZipPassword.NewPassword.Length > 0 ? true : false) : true;

        /// <summary>
        /// True wenn ein Password verwendet werden soll
        /// </summary>
        public bool UseEnryption { get; set; }

        /// <summary>
        /// True wenn Files in der Auflistung vorhanden sind
        /// </summary>
        public bool HasItems => Items.Count > 0;

        /// <summary>
        /// die Anzahl der Files in der Auflistung
        /// </summary>
        public int ItemsCount => Items.Count;

        #endregion

        #region Commands

        /// <summary>
        /// der Command zum Erstellen des Zip
        /// </summary>
        public ICommand CreateZipCommand { get; set; }

        /// <summary>
        /// der Command zum leeren der gesamten Auflistung
        /// </summary>
        public ICommand ClearItemsCommand { get; set; }

        /// <summary>
        /// der Command um die Einstellungen aufzurufen
        /// </summary>
        public ICommand ShowSettingsCommand { get; set; }

        public ICommand ShowAboutCommand { get; set; }

        /// <summary>
        /// der Command um ein einzelnes File aus der Auflistung zu löschen
        /// </summary>
        public ICommand DeleteItemCommand { get; set; }

        /// <summary>
        /// der Command um ein Passwort setzen zu können
        /// </summary>
        public ICommand SetEncryptionPasswordCommand { get; set; }
        #endregion

        #region Command Methods

        /// <summary>
        /// die Methode um ein Zip Archiv zu erstellen.
        /// damit dem User ein Hinweis angezeigt werden kann, ist die Methode asyncron
        /// </summary>
        /// <returns></returns>
        private async Task CreateZipAsync()
        {
            // Asynchronen Command einleiten
            // Viewmodel Bescheid geben das eine Aktion läuft
            await RunCommandAsync(() => VmIsBusy, async () =>
            {
                // wenn ein Passwort gesetzt werden soll
                if (UseEnryption)
                {
                    //Auflistung durchgehen und der Methode zum zippen übergeben und Passwort übergeben
                    await Items.Select(f => f.FileName).CreateEncryptetZipFileAsync(ZipPassword.NewPassword);
                    UseEnryption = false;
                }
                else
                {
                    //Auflistung durchgehen und der Methode zum zippen übergeben ohne Passwort
                    await Items.Select(f => f.FileName).CreateZipFileAsync();
                }
                //Meldung ausgeben das wir fertig sind
                await IoC.UI.ShowMessage(new DialogMessageBoxViewModel
                {
                    Title = "Dateien gezippt...",
                    Message = "Dateien erfolgreich gezippt."
                }) ;
                ///Auflistung leeren
                ClearItems();
            });
        }
       
        /// <summary>
        /// die Methode um die Eingabe eines Passwortes einzuleiten
        /// </summary>
        private void SetEncryptionPassword()
        {
            ZipPassword = new PasswordBoxViewModel
            {
                FakePassword = "Kein Passwort gesetzt",
                NewPassword = new SecureString()
            };
        }

        /// <summary>
        /// Methode um die Auflistung zu leeren,
        /// vorsichtshalber die UseEncryption auf false setzen
        /// </summary>
        private void ClearItems()
        {
            Items = new ObservableCollection<DropFileViewModel>();
            UseEnryption = false;
        }

        /// <summary>
        /// Methode um ein einzelnes File aus der Auflistung zu löschen
        /// </summary>
        /// <param name="param"></param>
        private void DeleteSelectedItem(object param)
        {
            SelectedItem = param as DropFileViewModel;
            Items.Remove(SelectedItem);
        }

        #endregion

        #region Helper Methods

        /// <summary>
        /// Alle Commands initialisieren
        /// </summary>
        private void InitializeCommands()
        {
            CreateZipCommand = new RelayCommand(async c => await CreateZipAsync(), c => CanCreateZip);
            ClearItemsCommand = new RelayCommand(c => ClearItems());
            ShowSettingsCommand = new RelayCommand(s => IoC.Application.SettingsMenuVisible = true);
            ShowAboutCommand = new RelayCommand(a => IoC.Application.AboutMenuVisible = true);
            DeleteItemCommand = new RelayCommand(d => DeleteSelectedItem(d));
            SetEncryptionPasswordCommand = new RelayCommand(p => SetEncryptionPassword());
        }

        /// <summary>
        /// Methode die beim droppen von Files ausgeführt wird
        /// </summary>
        /// <param name="files"></param>
        public void OnFileDrop(IEnumerable<string> files)
        {
            // neue Auflistung erstellen und alle gedroppten Files hinzufügen
            Items = new ObservableCollection<DropFileViewModel>(files.Select(f => new DropFileViewModel(f)));
            // auf Änderungen reagieren einschalten
            Items.CollectionChanged += Items_CollectionChanged;
            // ausgewähltes Item setzen
            SelectedItem = Items.FirstOrDefault();
        }

        /// <summary>
        /// Methode für die Änderungen an der Liste
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Items_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            // die beiden Properties updaten
            OnPropertyChanged(nameof(HasItems));
            OnPropertyChanged(nameof(ItemsCount));
        }

        #endregion
    }
}
