﻿using System;
using System.Diagnostics;
using System.Windows.Input;

namespace de.mhoapps.DropToZip.ViewModel
{
    [Serializable]
    /// <summary>
    /// Basisklasse für Commands in WPF
    /// </summary>
    public class RelayCommand : ICommand
    {
        #region Private Members


        private readonly Action<object> _execute;
        private readonly Predicate<object> _canExecute;

        #endregion

        #region Public Events

        /// <summary>
        /// das Event, was ausgelöst wird, wenn der <see cref="CanExecute(object)"/> Wert sich geändert hat
        /// </summary>
        public event EventHandler CanExecuteChanged
        {
            add
            {
                if (_canExecute != null)
                    CommandManager.RequerySuggested += value;
            }
            remove
            {
                if (_canExecute != null)
                    CommandManager.RequerySuggested -= value;
            }
        }
        void OnCanExecuteChanged(object sender, EventArgs e)
        {
        }

        #endregion

        #region Constructor

        /// <summary>
        /// Constructor für einen Command, der NUR executet werden kann
        /// </summary>
        /// <param name="execute">die logik für execute</param>
        public RelayCommand(Action<object> execute) : this(execute, null)
        { }

        /// <summary>
        /// constructor für einen command der sowohl Logik für execute, als auch für canexecute enthält
        /// </summary>
        /// <param name="execute"> die Logik für execute</param>
        /// <param name="canExecute"> die logik für canexecute</param>
        public RelayCommand(Action<object> execute, Predicate<object> canExecute)
        {
            if (execute == null)
                throw new ArgumentNullException("execute");
            _execute = execute;
            _canExecute = canExecute;
        }

        #endregion

        #region Command Methods

        /// <summary>
        /// setzt die CanExecute-Methode des Icommand-Interface auf true oder false
        /// </summary>
        /// <param name="parameter"></param>
        /// <returns>gibt zurück, ob die Aktion ausgeführt werden kann oder nicht</returns>
        /// <remarks>der Debugger überspringt diese Prozedur, es sei denn es wird ein Haltepunkt gesetzt</remarks>
        [DebuggerStepThrough]
        public bool CanExecute(object parameter)
        {
            return _canExecute == null || _canExecute(parameter);
        }



        /// <summary>
        /// führt die Aktion des Commands aus
        /// </summary>
        /// <param name="parameter"></param>
        public void Execute(object parameter)
        {
            _execute(parameter);
        }

        #endregion
    }
}
