﻿using System.Security;
using System.Windows.Input;

namespace de.mhoapps.DropToZip.ViewModel
{
    /// <summary>
    /// ViewModel für die PasswordBox
    /// </summary>
    public class PasswordBoxViewModel : ViewModelBase
    {
        #region Public Properties

        /// <summary>
        /// dient zur Anzeige des Passwortes, oder eines Fake String
        /// </summary>
        public string FakePassword { get; set; }

        /// <summary>
        /// Das Password, welches eingegeben wird
        /// </summary>
        public SecureString NewPassword { get; set; }

        /// <summary>
        /// True wenn der Bearbeitungsmodes eingeschaltet wird
        /// </summary>
        public bool Editing { get; set; }

        /// <summary>
        /// True wenn das Passwort in Klarschrift angezeigt werden soll
        /// </summary>
        public bool ShowPassword { get; set; }

        /// <summary>
        /// true wenn ein Password vorhanden ist
        /// </summary>
        public bool HasPassword => NewPassword.Length > 0;

        #endregion

        #region Public Commands
        /// <summary>
        /// Command für den Bearbeitungsmodus
        /// </summary>
        public ICommand EditPasswordCommand { get; set; }
        /// <summary>
        /// Command für das Speichern des Passwortes
        /// </summary>
        public ICommand SavePasswordCommand { get; set; }

        /// <summary>
        /// Command um den Bearbeitungsmodus zu beenden
        /// </summary>
        public ICommand CancelCommand { get; set; }

        /// <summary>
        /// Command um das Password in Klarschrift umzuwandeln
        /// </summary>
        public ICommand UnsecurePasswordCommand { get; set; }

        #endregion

        #region Constructor 

        /// <summary>
        /// Default constructor
        /// </summary>
        public PasswordBoxViewModel()
        {
            InitializeCommands();
            if (IsInDesignMode)
            {
                FakePassword = "********";
            }
        }

        #endregion

        #region Command Methods

        /// <summary>
        /// Methode um den Bearbeitungsmodus einzuschalten
        /// </summary>
        private void EditPassword()
        {
            Editing = true;
            ShowPassword = false;
        }

        /// <summary>
        /// Methode um den Bearbeitungsmodus zu beenden
        /// </summary>
        private void Cancel()
        {
            // NewPassword zurücksetzen damit keine Passwörter in der Anwendung rum liegen
            NewPassword = new SecureString();
            FakePassword = "Kein Passwort gesetzt";
            Editing = false;
        }

        /// <summary>
        /// Methode um das sichere Password anzuzeigen
        /// </summary>
        private void UnsecurePassword()
        {
            if (ShowPassword)
                FakePassword = NewPassword.Unsecure();
            else
                FakePassword = "********";
        }

        /// <summary>
        /// das sichere Passwort liegt in NewPassword (SecureString)
        /// und wird im MainViewModel beim zippen verwendet
        /// es wird nirgendwo gespeichert!!
        /// </summary>
        private void SavePassword()
        { 
            FakePassword = "*******";
            Editing = false;
            // Meldung ausgeben das das Passwort nun verfügbar ist
            // siehe MainViewModel -> CreateZipAsync
            IoC.UI.ShowMessage(new DialogMessageBoxViewModel
            {
                Title = "Passwort gesetzt...",
                Message = "Passwort wurde für diesen Container gesetzt !"
            });
        }

        #endregion

        #region Helper Methods
        /// <summary>
        /// Methode um alle Commands zu initialisieren
        /// </summary>
        private void InitializeCommands()
        {
            EditPasswordCommand = new RelayCommand(e => EditPassword());
            CancelCommand = new RelayCommand(c => Cancel());
            SavePasswordCommand = new RelayCommand(s => SavePassword(), s => HasPassword);
            UnsecurePasswordCommand = new RelayCommand(u => UnsecurePassword(), c => HasPassword);
        }

        #endregion
    }
}
